#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/select.h>

#define NUM_THREADS 100
#define BUFSIZE 4096

struct sys_command_struct {
	char *cmd;
	char *output;
	int ret, len;
	int buf_mult;
	FILE *fp;
	int fd;
};

int fd_start_cmd(struct sys_command_struct *);

int main()
{
	struct sys_command_struct scmd[NUM_THREADS];
	fd_set rs;
	int high_fd = 0, i, result, nfound;
	struct timeval to = { 1, 0 };

	FD_ZERO(&rs);
	for(i = 0; i < NUM_THREADS; i++) {
		memset(&scmd[i], 0, sizeof(struct sys_command_struct));
		scmd[i].cmd = "/bin/ls -l /etc/{passwd,shadow}";
		result = fd_start_cmd(&scmd[i]);
		if(result == -1) {
			printf("fd_start_cmd returned -1!\n");
			continue;
		}

		/* add filedescriptor to poll for input */
		FD_SET(scmd[i].fd, &rs);
		if(scmd[i].fd >= high_fd) high_fd = scmd[i].fd;
	}
	while((nfound = select(high_fd + 1, &rs, NULL, NULL, &to)) != 0) {
		if(nfound == -1) {
			perror("select");
			exit(1);
		}
		printf("Input ready on %d descriptors\n", nfound);
		for(i = 0; i < NUM_THREADS; i++) {
			if(FD_ISSET(scmd[i].fd, &rs)) {
				printf("%d: examining EOF condition\n", i);

				/* check for EOF on the filepointer.
				 * If found, pclose and remove the descriptor from the
				 * fd_set. high_fd should really point to highest descriptor
				 * always, but we just decrement by one. */
				if(feof(scmd[i].fp)) {
					scmd[i].ret = pclose(scmd[i].fp);
					FD_CLR(scmd[i].fd, &rs);
					printf("Descriptor %d is finished (return: %d)\n",
						   i, scmd[i].ret);
					continue;
				}
				printf("Input ready on descriptor %d\n", i);
				/* check for overflow and adjust buffer-size accordingly */
				if(!(scmd[i].len % BUFSIZE))
					scmd[i].output = realloc(scmd[i].output, scmd[i].len + BUFSIZE);
				/* read the pending input and adjust read size counter */
				fgets(&scmd[i].output[scmd[i].len],
					  BUFSIZE - (scmd[i].len % BUFSIZE),
					  scmd[i].fp);
				scmd[i].len = strlen(scmd[i].output);
				printf("%d buf is %d bytes\n", i, scmd[i].len);
			}
		}
	}

	for(i = 0; i < NUM_THREADS; i++) {
		printf("Command: %s\n", scmd[i].cmd);
		printf("Return: %d\n", scmd[i].ret);
		printf("Output (%d bytes)::\n", scmd[i].len);
		printf("%s\n", scmd[i].output ? scmd[i].output : "(NO output!)");
	}

	return 0;
}

int fd_start_cmd(struct sys_command_struct *cs)
{
	cs->fp = popen(cs->cmd, "r");
	if(cs->fp == NULL) return -1;

	cs->fd = cs->fp->_fileno;
	return cs->fd;
}
